"""Apply SVD."""
import logging
from scipy.sparse.linalg import svds

logger = logging.getLogger(__name__)

__all__ = ('apply_sparse_svd')


# pylint: disable=C0103
def apply_sparse_svd(M, dim):
    """Apply SVD to sparse CSR matrix."""
    if dim == 0 or dim >= min(M.shape):
        logger.warning('Specified k={} null or exceeds matrix shape limit = '
                       '{}. Resetting k to {}'.format(dim, min(M.shape),
                                                      min(M.shape) - 1))
        dim = min(M.shape) - 1
    logger.info('Running SVD...')
    U, S, _ = svds(M, k=dim, which='LM', return_singular_vectors='u')
    S = S[::-1]  # put singular values in decreasing order of values
    U = U[:, ::-1]  # put singular vectors in decreasing order of sing. values
    return U, S



"""Welcome to counterix.

This is the entry point of the application.
"""
import os

import argparse
import logging
import logging.config

import embeddix

import counterix.core.generator as generator
import counterix.core.weigher as weigher
import counterix.core.reducer as reducer
import counterix.utils.config as cutils

logging.config.dictConfig(
    cutils.load(
        os.path.join(os.path.dirname(__file__), 'logging', 'logging.yml')))

logger = logging.getLogger(__name__)


def _generate(corpus_filepath, min_count, win_size):
    logger.info('Generating distributional model from {}'
                .format(corpus_filepath))
    if not corpus_filepath.endswith('.txt'):
        model_filepath = os.path.abspath(corpus_filepath)
    else:
        model_filepath = os.path.abspath(corpus_filepath).split('.txt')[0]
    model_filepath = '{}.mc-{}.win-{}'.format(model_filepath, min_count,
                                              win_size)
    vocab_filepath = '{}.vocab'.format(model_filepath)
    model, vocab = generator.generate_raw_count_based_dsm(
        corpus_filepath, min_count, win_size)
    logger.info('Saving vocabulary to file...')
    embeddix.save_vocab(vocab_filepath, vocab)
    logger.info('Saving raw count sparse matrix to file...')
    embeddix.save_sparse(model_filepath, model)


def generate(args):
    """Generate raw count model in scipy sparse CSR matrix format."""
    _generate(args.corpus, args.min_count, args.win_size)


def _weigh(model_filepath):
    logger.info('Applying PPMI-weighing to model {}'.format(model_filepath))
    model = embeddix.load_sparse(model_filepath)
    output_filepath = '{}.ppmi'.format(
        os.path.abspath(model_filepath).split('.npz')[0])
    ppmi_weighed_matrix = weigher.weigh(model)
    logger.info('Saving PPMI sparse matrix to file...')
    embeddix.save_sparse(output_filepath, ppmi_weighed_matrix)


def weigh(args):
    """Apply PPMI-weighing to input model."""
    _weigh(args.model)


# pylint: disable=C0103
def _svd(model_filepath, k):
    logger.info('Applying SVD with k={} to model {}'.format(k, model_filepath))
    sing_vectors_filepath = '{}.k-{}.singvectors'.format(
        os.path.abspath(model_filepath).split('.npz')[0], k)
    sing_values_filepath = '{}.k-{}.singvalues'.format(
        os.path.abspath(model_filepath).split('.npz')[0], k)
    model = embeddix.load_sparse(model_filepath)
    sing_vectors, sing_values = reducer.apply_sparse_svd(model, k)
    logger.info('Saving singular vectors to file...')
    embeddix.save_dense(sing_vectors_filepath, sing_vectors)
    logger.info('Saving singular values to file...')
    embeddix.save_dense(sing_values_filepath, sing_values)


def svd(args):
    """Apply SVD."""
    _svd(args.model, args.dim)


def main():
    """Launch counterix."""
    parser = argparse.ArgumentParser(prog='counterix')
    subparsers = parser.add_subparsers()
    parser_generate = subparsers.add_parser(
        'generate', formatter_class=argparse.RawTextHelpFormatter,
        help='generate raw frequency count based model')
    parser_generate.set_defaults(func=generate)
    parser_generate.add_argument('-c', '--corpus', required=True,
                                 help='absolute filepath to corpus .txt file')
    parser_generate.add_argument('-m', '--min-count', default=0, type=int,
                                 help='frequency threshold on vocabulary')
    parser_generate.add_argument('-w', '--win-size', default=2, type=int,
                                 help='size of context window')
    parser_weigh = subparsers.add_parser(
        'weigh', formatter_class=argparse.RawTextHelpFormatter,
        help='apply PPMI weighing to input sparse CSR matrix')
    parser_weigh.set_defaults(func=weigh)
    parser_weigh.add_argument('-m', '--model', required=True,
                              help='absolute path to .npz matrix '
                              'corresponding to the distributional '
                              'space to weigh')
    parser_svd = subparsers.add_parser(
        'svd', formatter_class=argparse.RawTextHelpFormatter,
        help='apply svd to input matrix')
    parser_svd.set_defaults(func=svd)
    parser_svd.add_argument('-m', '--model', required=True,
                            help='absolute path to .npz matrix '
                                 'corresponding to the distributional '
                                 'space to reduce via svd')
    parser_svd.add_argument('-k', '--dim', default=0, type=int,
                            help='number of dimensions in final model')
    args = parser.parse_args()
    args.func(args)



"""A basic count-based model using sparse matrices and no ppmi."""
import logging
from collections import defaultdict

from scipy import sparse
from tqdm import tqdm

import embeddix

logger = logging.getLogger(__name__)

__all__ = ('generate_raw_count_based_dsm')


def count_words(corpus_filepath, min_count):
    """Count words in a corpus."""
    _counts = defaultdict(int)
    logger.info('Counting words in {}'.format(corpus_filepath))
    with open(corpus_filepath, 'r', encoding='utf-8') as input_stream:
        for line in input_stream:
            line = line.strip()
            for word in line.split():
                _counts[word] += 1
    if min_count == 0:
        counts = _counts
    else:
        counts = {word: count for word, count in _counts.items()
                  if count >= min_count}
        logger.info('Filtering out vocabulary words with counts lower '
                    'than {}, shrinking size by {:.2f}% from {} to {}.'
                    .format(min_count, 100-len(counts)*100.0/len(_counts),
                            len(_counts), len(counts)))
    return counts


def _count_raw_no_filter(corpus_filepath, win_size, word_to_idx_dic,
                         total_num_lines):
    data_dic = defaultdict(lambda: defaultdict(int))
    with open(corpus_filepath, 'r', encoding='utf-8') as input_stream:
        for line in tqdm(input_stream, total=total_num_lines):
            tokens = line.strip().split()
            # raw count with symmetric matrix
            for token_pos, token in enumerate(tokens):
                start = 0 if win_size == 0 else max(0, token_pos-win_size)
                while start < token_pos:
                    ctx = tokens[start]
                    if token in word_to_idx_dic and ctx in word_to_idx_dic:
                        token_idx = word_to_idx_dic[token]
                        ctx_idx = word_to_idx_dic[ctx]
                        data_dic[token_idx][ctx_idx] += 1
                    start += 1
        return data_dic


def generate_raw_count_based_dsm(corpus_filepath, min_count, win_size):
    """Generate a count-based distributional model with raw counts."""
    word_to_count_dic = count_words(corpus_filepath=corpus_filepath,
                                    min_count=min_count)
    word_to_idx_dic = {word: idx for idx, word
                       in enumerate(word_to_count_dic.keys())}
    total_num_lines = embeddix.count_lines(corpus_filepath)
    data_dic = _count_raw_no_filter(corpus_filepath, win_size, word_to_idx_dic,
                                    total_num_lines)
    logger.info('Building CSR sparse matrix...')
    rows = []
    columns = []
    data = []
    for row_idx in tqdm(data_dic):
        for col_idx in data_dic[row_idx]:
            rows.append(row_idx)
            columns.append(col_idx)
            data.append(data_dic[row_idx][col_idx])
            # rely on the fact that the matrix is symmetric
            rows.append(col_idx)
            columns.append(row_idx)
            data.append(data_dic[row_idx][col_idx])
    model = sparse.csr_matrix((data, (rows, columns)),
                              shape=(len(word_to_idx_dic),
                                     len(word_to_idx_dic)),
                              dtype='f')
    # logger.info('Matrix info: {} non-zero entries, {} shape, {:.6f} density'
    #             .format(model.getnnz(), model.shape,
    #                     model.getnnz()*1.0/(model.shape[0]*model.shape[1])))
    return model, word_to_idx_dic



"""PPMI weighing func."""

import logging
import numpy as np
import scipy.sparse as sparse

logger = logging.getLogger(__name__)

__all__ = ('weigh')


def weigh(csr_matrix):
    """Return a ppmi-weighted CSR sparse matrix from an input CSR matrix."""
    logger.info('Weighing raw count CSR matrix via PPMI')
    words = sparse.csr_matrix(csr_matrix.sum(axis=1))
    contexts = sparse.csr_matrix(csr_matrix.sum(axis=0))
    total_sum = csr_matrix.sum()
    # csr_matrix = csr_matrix.multiply(words.power(-1)) # #(w, c) / #w
    # csr_matrix = csr_matrix.multiply(contexts.power(-1))  # #(w, c) / (#w * #c)
    # csr_matrix = csr_matrix.multiply(total)  # #(w, c) * D / (#w * #c)
    csr_matrix = csr_matrix.multiply(words.power(-1))\
                           .multiply(contexts.power(-1))\
                           .multiply(total_sum)
    csr_matrix.data = np.log2(csr_matrix.data)  # PMI = log(#(w, c) * D / (#w * #c))
    csr_matrix = csr_matrix.multiply(csr_matrix > 0)  # PPMI
    csr_matrix.eliminate_zeros()
    return csr_matrix
